/* sgmlio.c -
   IO functions for core parser.

   Written by James Clark (jjc@jclark.com).
*/

/* SGML must see a file in which records start with RS and end with
   RE, and EOFCHAR (Ctl-Z) is present at the end.  This module must
   supply these characters if they are not naturally present in the
   file.  SGML will open two files at a time: when an entity is
   nested, the new file is opened before closing the old in order to
   make sure the open is successful. If it is, the original open file
   is closed temporarily (IOPEND); when the stack is popped, the new
   file is closed and the original file is re-opened (IOCONT). SGML
   will check error returns for the initial open of a file and all
   reads, and for re-openings when the stack is popped, but not for
   closes.  Returning <0 indicates an error; 0 or more is a successful
   operation, except for IOREAD where the return value is the number
   of characters read, and must exceed 0 to be successful.  The first
   READ must always be successful, and normally consists of just
   priming the buffer with EOBCHAR (or RS EOBCHAR).  SGMLIO must
   assure that there is an EOBCHAR at the end of each block read,
   except for the last block of the entity, which must have an
   EOFCHAR.

   SGML views an entity as a contiguous whole, without regard to its
   actual form of storage.  SGMLIO supports entities that are
   equivalent to a single file of one or more records, or to a
   concatenation of files.
*/

/* Uses only stream I/O.  This module should be portable to most ANSI
   systems. */
/* We try to ensure that if an IO operation fails, then errno will contain
   a meaningful value (although it may be zero.) */

#include "config.h"
#ifdef HAVE_O_NOINHERIT
#include <fcntl.h>
#include <io.h>
#endif /* HAVE_O_NOINHERIT */

#include "sgmlaux.h"          /* Include files for auxiliary functions.. */

#ifdef HAVE_O_NOINHERIT
#define FOPENR(file) nifopen(file)
FILE *nifopen P((char *));
#else /* not HAVE_O_NOINHERIT */
#define FOPENR(file) fopen((file), "r")
#endif /* not HAVE_O_NOINHERIT */

struct iofcb {                /* I/O file control block. */
     FILE *fp;		      /* File handle. */
     long off;                /* Offset in file of current read block. */
     char *next;              /* Next file (NULL if no more). */
     char *file;              /* Current file (no length byte). */
     int sverrno;	      /* Saved value of errno. */
     char bol;	              /* Non-zero if currently at beginning of line. */
     char first;	      /* Non-zero if the first read.  */
     char wasbol;	      /* Non-zero if current block was at beginning of line. */
};

static char *lastfile;	      /* The name of the last file closed. */
static int bufsize;	      /* Size of buffer passed to ioread(). */
static char isbad[256];	      /* Table of chars that will be turned into BADCHAR. */

static char *nextstr();	      /* Iterate over list of strings. */

VOID ioinit(swp)
struct switches *swp;
{
     isbad[BADCHAR] = 1;
     isbad[RECHAR] = 1;
     isbad[EOBCHAR] = 1;
     isbad[EOFCHAR] = 1;
     isbad[EOS] = 1;
     isbad[(UNCH)swp->delnonch] = 1;
     isbad[(UNCH)GENRECHAR] = 1;
     bufsize = swp->swbufsz;
}

int ioopen(id, pp)
UNIV id;
UNIV *pp;
{
     struct iofcb *f;
     char *s;
     errno = 0;
     if (!id)
	  return -1;
     s = id;
     if (!*s)
	  return -1;
     f = (struct iofcb *)rmalloc((UNS)sizeof(struct iofcb));
     f->file = s;
     f->next = nextstr(s);
     errno = 0;
     f->fp = FOPENR(f->file);
     f->bol = 1;
     f->first = 1;
     *pp = (UNIV)f;
     return f->fp ? 1 : -1;
}

VOID ioclose(p)
UNIV p;
{
     struct iofcb *f = (struct iofcb *)p;
     fclose(f->fp);
     lastfile = f->file;
     frem((UNIV)f);
}

VOID iopend(p, off)
UNIV p;
int off;
{
     struct iofcb *f = (struct iofcb *)p;
     f->bol = 0;
     if (f->wasbol) {
	  if (off == 0)
	       f->bol = 1;
	  else
	       off--;
     }
     errno = 0;
     /* f->off will be checked in iocont. */
     if (fseek(f->fp, f->off, SEEK_SET))
	  f->off = -1L;
     else {
	  /* Avoid relying on position arithmetic. */
	  int c = EOF;
	  while (--off >= 0)
	       c = getc(f->fp);
	  f->off = ftell(f->fp);                  /* Save location. */
	  if (c == '\n')
	       f->bol = 1;
     }
     fclose(f->fp);
     f->sverrno = errno;
}

int iocont(p)
UNIV p;
{
     struct iofcb *f = (struct iofcb *)p;
     if (f->off == -1L) {
	  errno = f->sverrno;
	  return -1;
     }
     errno = 0;
     f->fp = FOPENR(f->file);
     if (f->fp) {
	  errno = 0;
	  return fseek(f->fp, f->off, SEEK_SET) ? -1 : 0;
     }
     else
	  return -1;
}

/* Return -1 on error, otherwise the number of bytes read.  The
strategy is to concatenate the files, insert a RS at the beginning of
each line, and change each '\n' into a RE.  The returned data
shouldn't cross a file boundary, otherwise error messages might be
inaccurate.  The first read must always succeed. */

int ioread(p, buf, newfilep)
UNIV p;
UNCH *buf;
int *newfilep;
{
     int i = 0;
     struct iofcb *f = (struct iofcb *)p;
     FILE *fp;
     int c;
     
     *newfilep = 0;
     if (f->first) {
	  buf[i] = EOBCHAR;
	  f->first = 0;
	  return 1;
     }
     fp = f->fp;
     for (;;) {
	  errno = 0;
	  f->off = ftell(fp);
	  if (f->off == -1L)
	       return -1;
	  errno = 0;
	  c = getc(fp);
	  if (c != EOF)
	       break;
	  if (ferror(fp))
	       return -1;
	  if (fclose(fp) == EOF)
	       return -1;
	  if (!f->next){
	       buf[0] = EOFCHAR;
	       return 1;
	  }
	  f->file = f->next;
	  f->next = nextstr(f->next);
	  *newfilep = 1;
	  errno = 0;
	  fp = f->fp = FOPENR(f->file);
	  if (!fp)
	       return -1;
	  f->bol = 1;
     }
     if (f->bol) {
	  f->bol = 0;
	  buf[i++] = RSCHAR;
	  f->wasbol = 1;
     }
     else
	  f->wasbol = 0;
     errno = 0;
     for (;;) {
	  if (c == '\n') {
	       f->bol = 1;
	       buf[i++] = RECHAR;
	       break;
	  }
	  if (isbad[c])
	       c = BADCHAR;
	  buf[i++] = c;
	  if (i >= bufsize - 1)
	       break;
	  c = getc(fp);
	  if (c == EOF) {
	       if (ferror(fp))
		    return -1;
	       /* This is in the middle of a line. */
	       break;
	  }
     }
     buf[i++] = EOBCHAR;
     return i;
}

static char *nextstr(p)
char *p;
{
     p = strchr(p, '\0');
     return *++p ? p : 0;
}

/* Return the filename associated with p.  If p is NULL, return the filename
of the last file closed. */

char *ioflid(p)
UNIV p;
{
     if (!p)
	  return lastfile;
     return ((struct iofcb *)p)->file;
}

#ifdef HAVE_O_NOINHERIT

/* This is the same as fopen(name, "r") except that it tells DOS that
the file descriptor should not be inherited by child processes.  */

FILE *nifopen(name)
char *name;
{
     int fd = open(name, O_RDONLY|O_NOINHERIT|O_TEXT);
     if (fd < 0)
	  return 0;
     return fdopen(fd, "r");
}

#endif /* HAVE_O_NOINHERIT */

/*
Local Variables:
c-indent-level: 5
c-continued-statement-offset: 5
c-brace-offset: -5
c-argdecl-indent: 0
c-label-offset: -5
End:
*/
